# flake8: noqa
from gevent import monkey

monkey.patch_all()


import asyncio
import tempfile
import shlex
import sys

from rbx.box.schema import CodeItem, ExpectedOutcome, TestcaseGroup


import pathlib
import shutil
from typing import Annotated, List, Optional

import rich
import rich.prompt
import typer
import questionary

from rbx import annotations, config, console, utils
from rbx.box import (
    builder,
    cd,
    setter_config,
    creation,
    download,
    environment,
    generators,
    package,
    compile,
    presets,
    stresses,
    validators,
)
from rbx.box.contest import main as contest
from rbx.box.environment import VerificationLevel, get_environment_path
from rbx.box.packaging import main as packaging
from rbx.box.solutions import (
    estimate_time_limit,
    get_exact_matching_solutions,
    get_matching_solutions,
    print_run_report,
    run_and_print_interactive_solutions,
    run_solutions,
)
from rbx.box.statements import build_statements

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)
app.add_typer(
    setter_config.app,
    name='config, cfg',
    cls=annotations.AliasGroup,
    help='Manage setter configuration.',
)
app.add_typer(
    build_statements.app,
    name='statements, st',
    cls=annotations.AliasGroup,
    help='Manage statements.',
)
app.add_typer(
    download.app,
    name='download',
    cls=annotations.AliasGroup,
    help='Download an asset from supported repositories.',
)
app.add_typer(
    presets.app, name='presets', cls=annotations.AliasGroup, help='Manage presets.'
)
app.add_typer(
    packaging.app,
    name='package, pkg',
    cls=annotations.AliasGroup,
    help='Build problem packages.',
)
app.add_typer(
    contest.app, name='contest', cls=annotations.AliasGroup, help='Contest management.'
)


# @app.command('ui', hidden=True)
# @package.within_problem
# def ui():
#     ui_pkg.start()


@app.command('edit, e', help='Open problem.rbx.yml in your default editor.')
@package.within_problem
def edit():
    console.console.print('Opening problem definition in editor...')
    # Call this function just to raise exception in case we're no in
    # a problem package.
    package.find_problem()
    config.open_editor(package.find_problem_yaml() or pathlib.Path())


@app.command('build, b', help='Build all tests for the problem.')
@package.within_problem
def build(verification: environment.VerificationParam):
    builder.build(verification=verification)


@app.command('run, r', help='Build and run solution(s).')
@package.within_problem
def run(
    verification: environment.VerificationParam,
    solution: Annotated[
        Optional[str],
        typer.Argument(
            help='Path to solution to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
    timeit: bool = typer.Option(
        False,
        '--time',
        '-t',
        help='Whether to use estimate a time limit based on accepted solutions.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
):
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    if not builder.build(verification=verification, output=check):
        return

    if verification <= VerificationLevel.VALIDATE.value:
        console.console.print(
            '[warning]Verification level is set to [item]validate (-v1)[/item], so rbx only build tests and validated them.[/warning]'
        )
        return

    override_tl = None
    if timeit:
        if sanitized:
            console.console.print(
                '[error]Sanitizers are known to be time-hungry, so they cannot be used for time estimation.\n'
                'Remove either the [item]-s[/item] flag or the [item]-t[/item] flag to run solutions without sanitizers.[/error]'
            )
            raise typer.Exit(1)

        # Never use sanitizers for time estimation.
        override_tl = _time_impl(check=check, detailed=False)
        if override_tl is None:
            raise typer.Exit(1)

    with utils.StatusProgress('Running solutions...') as s:
        if sanitized:
            console.console.print(
                '[warning]Sanitizers are running, so the time limit for the problem will be dropped, '
                'and the environment default time limit will be used instead.[/warning]'
            )

        tracked_solutions = None
        if outcome is not None:
            tracked_solutions = {
                str(solution.path)
                for solution in get_matching_solutions(ExpectedOutcome(outcome))
            }
        if solution:
            tracked_solutions = {solution}

        if sanitized and tracked_solutions is None:
            console.console.print(
                '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
            )
            tracked_solutions = {
                str(solution.path)
                for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
            }

        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            timelimit_override=override_tl,
            sanitized=sanitized,
        )

    console.console.print()
    console.console.rule('[status]Run report[/status]', style='status')
    asyncio.run(
        print_run_report(
            solution_result,
            console.console,
            verification,
            detailed=detailed,
        )
    )


def _time_impl(check: bool, detailed: bool) -> Optional[int]:
    if package.get_main_solution() is None:
        console.console.print(
            '[warning]No main solution found, so cannot estimate a time limit.[/warning]'
        )
        return None

    verification = VerificationLevel.ALL_SOLUTIONS.value

    with utils.StatusProgress('Running ACCEPTED solutions...') as s:
        tracked_solutions = {
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        }
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
        )

    console.console.print()
    console.console.rule(
        '[status]Run report (for time estimation)[/status]', style='status'
    )
    ok = asyncio.run(
        print_run_report(
            solution_result,
            console.console,
            verification,
            detailed=detailed,
        )
    )

    if not ok:
        console.console.print(
            '[error]Failed to run ACCEPTED solutions, so cannot estimate a reliable time limit.[/error]'
        )
        return None

    console.console.print()
    return asyncio.run(estimate_time_limit(console.console, solution_result))


@app.command(
    'time, t',
    help='Estimate a time limit for the problem based on a time limit formula and timings of accepted solutions.',
)
@package.within_problem
def time(
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    detailed: bool = typer.Option(
        False,
        '--detailed',
        '-d',
        help='Whether to print a detailed view of the tests using tables.',
    ),
):
    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    verification = VerificationLevel.ALL_SOLUTIONS.value
    if not builder.build(verification=verification, output=check):
        return None

    _time_impl(check, detailed)


@app.command(
    'irun, ir', help='Build and run solution(s) by passing testcases in the CLI.'
)
@package.within_problem
def irun(
    verification: environment.VerificationParam,
    solution: Annotated[
        Optional[str],
        typer.Argument(
            help='Path to solution to run. If not specified, will run all solutions.'
        ),
    ] = None,
    outcome: Optional[str] = typer.Option(
        None,
        '--outcome',
        '-o',
        help='Include only solutions whose expected outcomes intersect with this.',
    ),
    check: bool = typer.Option(
        True,
        '--nocheck',
        flag_value=False,
        help='Whether to not build outputs for tests and run checker.',
    ),
    generator: Optional[str] = typer.Option(
        None,
        '--generator',
        '-g',
        help='Generator call to use to generate a single test for execution.',
    ),
    print: bool = typer.Option(
        False, '--print', '-p', help='Whether to print outputs to terminal.'
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
):
    if not print:
        console.console.print(
            '[warning]Outputs will be written to files. If you wish to print them to the terminal, use the "-p" parameter.'
        )
    if verification < VerificationLevel.ALL_SOLUTIONS.value:
        console.console.print(
            '[warning]Verification level should be at least [item]all solutions (-v4)[/item] to run solutions interactively.'
        )
        return

    main_solution = package.get_main_solution()
    if check and main_solution is None:
        console.console.print(
            '[warning]No main solution found, running without checkers.[/warning]'
        )
        check = False

    tracked_solutions = None
    if outcome is not None:
        tracked_solutions = {
            str(solution.path)
            for solution in get_matching_solutions(ExpectedOutcome(outcome))
        }
    if solution:
        tracked_solutions = {solution}
    if sanitized and tracked_solutions is None:
        console.console.print(
            '[warning]Sanitizers are running, and no solutions were specified to run. Will only run [item]ACCEPTED[/item] solutions.'
        )
        tracked_solutions = {
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        }

    asyncio.run(
        run_and_print_interactive_solutions(
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            generator=generators.get_call_from_string(generator)
            if generator is not None
            else None,
            print=print,
            sanitized=sanitized,
        )
    )


@app.command('create, c', help='Create a new problem package.')
def create(
    name: str,
    preset: Annotated[
        Optional[str], typer.Option(help='Preset to use when creating the problem.')
    ] = None,
):
    if preset is not None:
        creation.create(name, preset=preset)
        return
    creation.create(name)


@app.command('stress', help='Run a stress test.')
@package.within_problem
def stress(
    name: Annotated[
        str,
        typer.Argument(
            help='Name of the stress test to run (specified in problem.rbx.yml), '
            'or the generator to run, in case -g is specified.'
        ),
    ],
    generator_args: Annotated[
        Optional[str],
        typer.Option(
            '--generator',
            '-g',
            help='Run generator [name] with these args.',
        ),
    ] = None,
    finder: Annotated[
        Optional[str],
        typer.Option(
            '--finder',
            '-f',
            help='Run a stress with this finder expression.',
        ),
    ] = None,
    timeout: Annotated[
        int,
        typer.Option(
            '--timeout',
            '--time',
            '-t',
            help='For how many seconds to run the stress test.',
        ),
    ] = 10,
    findings: Annotated[
        int,
        typer.Option('--findings', '-n', help='How many breaking tests to look for.'),
    ] = 1,
    verbose: bool = typer.Option(
        False,
        '-v',
        '--verbose',
        help='Whether to print verbose output for checkers and finders.',
    ),
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the solutions with sanitizers enabled.',
    ),
):
    if finder and not generator_args or generator_args and not finder:
        console.console.print(
            '[error]Options --generator/-g and --finder/-f should be specified together.'
        )
        raise typer.Exit(1)

    with utils.StatusProgress('Running stress...') as s:
        report = stresses.run_stress(
            name,
            timeout,
            args=generator_args,
            finder=finder,
            findingsLimit=findings,
            progress=s,
            verbose=verbose,
            sanitized=sanitized,
        )

    stresses.print_stress_report(report)

    if not report.findings:
        return

    # Add found tests.
    res = rich.prompt.Confirm.ask(
        'Do you want to add the tests that were found to a test group?',
        console=console.console,
    )
    if not res:
        return
    testgroup = None
    while testgroup is None or testgroup:
        groups_by_name = {
            name: group
            for name, group in package.get_test_groups_by_name().items()
            if group.generatorScript is not None
            and group.generatorScript.path.suffix == '.txt'
        }

        testgroup = questionary.select(
            'Choose the testgroup to add the tests to.\nOnly test groups that have a .txt generatorScript are shown below: ',
            choices=list(groups_by_name) + ['(create new script)', '(skip)'],
        ).ask()

        if testgroup == '(create new script)':
            new_script_name = questionary.text(
                'Enter the name of the new .txt generatorScript file: '
            ).ask()
            new_script_path = pathlib.Path(new_script_name).with_suffix('.txt')
            new_script_path.parent.mkdir(parents=True, exist_ok=True)
            new_script_path.touch()

            # Temporarily create a new testgroup with the new script.
            testgroup = new_script_path.stem
            groups_by_name[testgroup] = TestcaseGroup(
                name=testgroup, generatorScript=CodeItem(path=new_script_path)
            )
            ru, problem_yml = package.get_ruyaml()
            problem_yml['testcases'].append(
                {
                    'name': testgroup,
                    'generatorScript': new_script_path.name,
                }
            )
            dest = package.find_problem_yaml()
            assert dest is not None
            utils.save_ruyaml(dest, ru, problem_yml)
            package.clear_package_cache()

        if testgroup not in groups_by_name:
            break
        try:
            subgroup = groups_by_name[testgroup]
            assert subgroup.generatorScript is not None
            generator_script = pathlib.Path(subgroup.generatorScript.path)

            finding_lines = []
            for finding in report.findings:
                line = finding.generator.name
                if finding.generator.args is not None:
                    line = f'{line} {finding.generator.args}'
                finding_lines.append(line)

            with generator_script.open('a') as f:
                stress_text = f'# Obtained by running `rbx {shlex.join(sys.argv[1:])}`'
                finding_text = '\n'.join(finding_lines)
                f.write(f'\n{stress_text}\n{finding_text}\n')

            console.console.print(
                f"Added [item]{len(report.findings)}[/item] tests to test group [item]{testgroup}[/item]'s generatorScript at [item]{subgroup.generatorScript.path}[/item]."
            )
        except typer.Exit:
            continue
        break


@app.command('compile', help='Compile an asset given its path.')
@package.within_problem
def compile_command(
    path: Annotated[str, typer.Argument(help='Path to the asset to compile.')],
    sanitized: bool = typer.Option(
        False,
        '--sanitized',
        '-s',
        help='Whether to compile the asset with sanitizers enabled.',
    ),
    warnings: bool = typer.Option(
        False,
        '--warnings',
        '-w',
        help='Whether to compile the asset with warnings enabled.',
    ),
):
    compile.any(path, sanitized, warnings)


@app.command('validate', help='Run the validator in a one-off fashion, interactively.')
@package.within_problem
def validate(
    path: Annotated[
        Optional[str],
        typer.Option('--path', '-p', help='Path to the testcase to validate.'),
    ] = None,
):
    validator_tuple = validators.compile_main_validator()
    if validator_tuple is None:
        console.console.print('[error]No validator found for this problem.[/error]')
        raise typer.Exit(1)

    validator, validator_digest = validator_tuple

    input = console.multiline_prompt('Testcase input')

    if path is None:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = pathlib.Path(tmpdir) / '000.in'
            tmppath.write_text(input)

            info = validators.validate_one_off(
                pathlib.Path(tmppath), validator, validator_digest
            )
    else:
        info = validators.validate_one_off(
            pathlib.Path(path), validator, validator_digest
        )

    validators.print_validation_report([info])


@app.command('environment, env', help='Set or show the current box environment.')
def environment_command(
    env: Annotated[Optional[str], typer.Argument()] = None,
    install_from: Annotated[
        Optional[str],
        typer.Option(
            '--install',
            '-i',
            help='Whether to install this environment from the given file.',
        ),
    ] = None,
):
    if env is None:
        cfg = config.get_config()
        console.console.print(f'Current environment: [item]{cfg.boxEnvironment}[/item]')
        console.console.print(
            f'Location: {environment.get_environment_path(cfg.boxEnvironment)}'
        )
        return
    if install_from is not None:
        environment.install_environment(env, pathlib.Path(install_from))
    if not get_environment_path(env).is_file():
        console.console.print(
            f'[error]Environment [item]{env}[/item] does not exist.[/error]'
        )
        raise typer.Exit(1)

    cfg = config.get_config()
    if env == cfg.boxEnvironment:
        console.console.print(
            f'Environment is already set to [item]{env}[/item].',
        )
        return
    console.console.print(
        f'Changing box environment from [item]{cfg.boxEnvironment}[/item] to [item]{env}[/item]...'
    )
    cfg.boxEnvironment = env
    config.save_config(cfg)

    # Also clear cache when changing environments.
    clear()


@app.command(
    'activate',
    help='Activate the environment of the current preset used by the package.',
)
@cd.within_closest_package
def activate():
    preset_lock = presets.get_preset_lock()
    if preset_lock is None:
        console.console.print(
            '[warning]No configured preset to be activated for this package.[/warning]'
        )
        raise typer.Exit(1)

    preset = presets.get_installed_preset_or_null(preset_lock.preset_name)
    if preset is None:
        if preset_lock.uri is None:
            console.console.print(
                '[error]Preset is not installed. Install it manually, or specify a URI in [item].preset-lock.yml[/item].[/error]'
            )
            raise typer.Exit(1)
        presets.install(preset_lock.uri)

    preset = presets.get_installed_preset(preset_lock.preset_name)

    # Install the environment from the preset if it's not already installed.
    presets.optionally_install_environment_from_preset(
        preset, root=presets.get_preset_installation_path(preset_lock.name)
    )

    # Activate the environment.
    if preset.env is not None:
        environment_command(preset.name)

    console.console.print(f'[success]Preset [item]{preset.name}[/item] is activated.')


@app.command('languages', help='List the languages available in this environment')
def languages():
    env = environment.get_environment()

    console.console.print(
        f'[success]There are [item]{len(env.languages)}[/item] language(s) available.'
    )

    for language in env.languages:
        console.console.print(
            f'[item]{language.name}[/item], aka [item]{language.readable_name or language.name}[/item]:'
        )
        console.console.print(language)
        console.console.print()


@app.command('clear, clean', help='Clears cache and build directories.')
@cd.within_closest_package
def clear():
    console.console.print('Cleaning cache and build directories...')
    shutil.rmtree('.box', ignore_errors=True)
    shutil.rmtree('build', ignore_errors=True)


@app.callback()
def callback():
    pass



from __future__ import annotations

import logging
import os
import pathlib
import shutil
import stat
import subprocess
import tempfile
from typing import IO, Any, Dict, List, Optional

from rbx.config import get_app_path
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.sandbox import (
    SandboxBase,
    SandboxParams,
    wait_without_std,
)

logger = logging.getLogger(__name__)


class IsolateSandbox(SandboxBase):
    """This class creates, deletes and manages the interaction with a
    sandbox. The sandbox doesn't support concurrent operation, not
    even for reading.

    The Sandbox offers API for retrieving and storing file, as well as
    executing programs in a controlled environment. There are anyway a
    few files reserved for use by the Sandbox itself:

     * commands.log: a text file with the commands ran into this
       Sandbox, one for each line;

     * run.log.N: for each N, the log produced by the sandbox when running
       command number N.

    """

    next_id = 0

    # If the command line starts with this command name, we are just
    # going to execute it without sandboxing, and with all permissions
    # on the current directory.
    SECURE_COMMANDS = ['/bin/cp', '/bin/mv', '/usr/bin/zip', '/usr/bin/unzip']

    log: Optional[Dict[str, Any]]

    def __init__(
        self,
        file_cacher: Optional[FileCacher] = None,
        name: Optional[str] = None,
        temp_dir: Optional[pathlib.Path] = None,
        params: Optional[SandboxParams] = None,
        debug: bool = False,
    ):
        """Initialization.

        For arguments documentation, see SandboxBase.__init__.

        """
        if not temp_dir:
            temp_dir = pathlib.Path(tempfile.gettempdir())
        SandboxBase.__init__(self, file_cacher, name, temp_dir, params)

        self.box_id = IsolateSandbox.next_id % 10
        IsolateSandbox.next_id += 1

        # We create a directory "home" inside the outer temporary directory,
        # that will be bind-mounted to "/tmp" inside the sandbox (some
        # compilers need "/tmp" to exist, and this is a cheap shortcut to
        # create it). The sandbox also runs code as a different user, and so
        # we need to ensure that they can read and write to the directory.
        # But we don't want everybody on the system to, which is why the
        # outer directory exists with no read permissions.
        self._outer_dir = pathlib.Path(
            tempfile.mkdtemp(dir=str(self.temp_dir), prefix='cms-%s-' % (self.name))
        )
        self._home = self._outer_dir / 'home'
        self._home_dest = pathlib.PosixPath('/tmp')
        self._home.mkdir(parents=True, exist_ok=True)
        self.allow_writing_all()

        self.exec_name = 'isolate'
        self.box_exec = self.detect_box_executable()
        # Used for -M - the meta file ends up in the outer directory. The
        # actual filename will be <info_basename>.<execution_number>.
        self.info_basename = self._outer_dir / 'run.log'
        self.log = None
        self.exec_num = -1
        self.cmd_file = self._outer_dir / 'commands.log'
        self.chdir = self._home_dest
        self.debug = debug
        logger.debug(
            "Sandbox in `%s' created, using box `%s'.", self._home, self.box_exec
        )

        # Ensure we add a few extra things to params.
        self.set_params(params or SandboxParams())

        # Tell isolate to get the sandbox ready. We do our best to cleanup
        # after ourselves, but we might have missed something if a previous
        # worker was interrupted in the middle of an execution, so we issue an
        # idempotent cleanup.
        self.cleanup()
        self.initialize()

    def set_params(self, params: SandboxParams):
        """Set the parameters of the sandbox.

        params (SandboxParams): the parameters to set.

        """
        super().set_params(params)
        self.add_mapped_directory(self._home, dest=self._home_dest, options='rw')

        # Set common environment variables.
        # Specifically needed by Python, that searches the home for
        # packages.
        self.params.set_env['HOME'] = str(self._home_dest)

    def add_mapped_directory(
        self,
        src: pathlib.Path,
        dest: Optional[pathlib.Path] = None,
        options: Optional[str] = None,
        ignore_if_not_existing: bool = False,
    ):
        """Add src to the directory to be mapped inside the sandbox.

        src (Path): directory to make visible.
        dest (Path|None): if not None, the path where to bind src.
        options (str|None): if not None, isolate's directory rule options.
        ignore_if_not_existing (bool): if True, ignore the mapping when src
            does not exist (instead of having isolate terminate with an
            error).

        """
        self.params.add_mapped_directory(
            src, dest, options, ignore_if_not_existing=ignore_if_not_existing
        )

    def maybe_add_mapped_directory(
        self,
        src: pathlib.Path,
        dest: Optional[pathlib.Path] = None,
        options: Optional[str] = None,
    ):
        """Same as add_mapped_directory, with ignore_if_not_existing."""
        return self.add_mapped_directory(
            src, dest, options, ignore_if_not_existing=True
        )

    def allow_writing_all(self):
        """Set permissions in such a way that any operation is allowed."""
        self._home.chmod(0o777)
        for child in self._home.iterdir():
            child.chmod(0o777)

    def allow_writing_none(self):
        """Set permissions in such a way that the user cannot write anything."""
        self._home.chmod(0o755)
        for child in self._home.iterdir():
            child.chmod(0o755)

    def allow_writing_only(self, inner_paths: List[pathlib.Path]):
        """Set permissions in so that the user can write only some paths.

        By default the user can only write to the home directory. This
        method further restricts permissions so that it can only write
        to some files inside the home directory.

        inner_paths ([Path]): the only paths that the user is allowed to
            write to; they should be "inner" paths (from the perspective
            of the sandboxed process, not of the host system); they can
            be absolute or relative (in which case they are interpreted
            relative to the home directory); paths that point to a file
            outside the home directory are ignored.

        """
        outer_paths: List[pathlib.Path] = []
        for inner_path in inner_paths:
            abs_inner_path = (self._home_dest / inner_path).resolve()
            # If an inner path is absolute (e.g., /fifo0/u0_to_m) then
            # it may be outside home and we should ignore it.
            if not abs_inner_path.is_relative_to(self._home_dest.resolve()):
                continue
            rel_inner_path = abs_inner_path.relative_to(self._home_dest)
            outer_path = self._home / rel_inner_path
            outer_paths.append(outer_path)

        # If one of the specified file do not exists, we touch it to
        # assign the correct permissions.
        for path in outer_paths:
            if not path.exists():
                path.touch()

        # Close everything, then open only the specified.
        self.allow_writing_none()
        for path in outer_paths:
            path.chmod(0o722)

    def get_root_path(self) -> pathlib.Path:
        """Return the toplevel path of the sandbox.

        return (Path): the root path.

        """
        return self._outer_dir

    def relative_path(self, path: pathlib.Path) -> pathlib.Path:
        """Translate from a relative path inside the sandbox to a system path.

        path (Path): relative path of the file inside the sandbox.

        return (Path): the absolute path.

        """
        return self._home / path

    def detect_box_executable(self) -> pathlib.Path:
        """Try to find an isolate executable. It first looks in
        ./isolate/, then the local directory, then in a relative path
        from the file that contains the Sandbox module, then in the
        system paths.

        return (Path): the path to a valid (hopefully) isolate.

        """
        paths: List[pathlib.Path] = [
            pathlib.PosixPath('./isolate') / self.exec_name,
            pathlib.PosixPath('.') / self.exec_name,
            get_app_path() / self.exec_name,
            pathlib.PosixPath('/usr/local/bin') / self.exec_name,
            pathlib.PosixPath(self.exec_name),
        ]
        for path in paths:
            # Consider only non-directory, executable files with SUID flag on.
            if path.exists() and not path.is_dir() and os.access(str(path), os.X_OK):
                st = path.stat()
                if st.st_mode & stat.S_ISUID != 0:
                    return path

        # As default, return self.exec_name alone, that means that
        # system path is used.
        return paths[-1]

    def build_box_options(self) -> List[str]:
        """Translate the options defined in the instance to a string
        that can be postponed to isolate as an arguments list.

        return ([string]): the arguments list as strings.

        """
        res = list()
        if self.box_id is not None:
            res += [f'--box-id={self.box_id}']
        if self.params.cgroup:
            res += ['--cg']
        if self.chdir is not None:
            res += [f'--chdir={str(self.chdir)}']
        for dirmount in self.params.dirs:
            s = str(dirmount.dst) + '=' + str(dirmount.src)
            if dirmount.options is not None:
                s += ':' + dirmount.options
            res += [f'--dir={s}']
        if self.params.preserve_env:
            res += ['--full-env']
        for var in self.params.inherit_env:
            res += [f'--env={var}']
        for var, value in self.params.set_env.items():
            res += [f'--env={var}={value}']
        if self.params.fsize is not None:
            # Isolate wants file size as KiB.
            fsize = self.params.fsize
            res += [f'--fsize={fsize}']
        if self.params.stdin_file is not None:
            inner_stdin = self.inner_absolute_path(self.params.stdin_file)
            res += ['--stdin=%s' % str(inner_stdin)]
        if self.params.stack_space is not None:
            # Isolate wants stack size as KiB.
            stack_space = self.params.stack_space * 1024
            res += [f'--stack={stack_space}']
        if self.params.address_space is not None:
            # Isolate wants memory size as KiB.
            address_space = self.params.address_space * 1024
            if self.params.cgroup:
                res += [f'--cg-mem={address_space}']
            else:
                res += [f'--mem={address_space}']
        if self.params.stdout_file is not None:
            inner_stdout = self.inner_absolute_path(self.params.stdout_file)
            res += ['--stdout=%s' % str(inner_stdout)]
        if self.params.max_processes is not None:
            res += [f'--processes={self.params.max_processes}']
        else:
            res += ['--processes']
        if self.params.stderr_file is not None:
            inner_stderr = self.inner_absolute_path(self.params.stderr_file)
            res += ['--stderr=%s' % str(inner_stderr)]
        if self.params.timeout is not None:
            # Isolate wants time in seconds.
            timeout = float(self.params.timeout) / 1000
            res += ['--time=%g' % timeout]
        res += ['--verbose'] * self.params.verbosity
        if self.params.wallclock_timeout is not None:
            wallclock_timeout = float(self.params.wallclock_timeout) / 1000
            res += ['--wall-time=%g' % wallclock_timeout]
        if self.params.extra_timeout is not None:
            extra_timeout = float(self.params.extra_timeout) / 1000
            res += ['--extra-time=%g' % extra_timeout]
        res += ['--meta=%s' % ('%s.%d' % (self.info_basename, self.exec_num))]
        res += ['--run']
        return res

    def hydrate_logs(self):
        """Read the content of the log file of the sandbox (usually
        run.log.N for some integer N), and set self.log as a dict
        containing the info in the log file (time, memory, status,
        ...).

        """
        # self.log is a dictionary of lists (usually lists of length
        # one).
        self.log = {}
        info_file = pathlib.Path('%s.%d' % (self.info_basename, self.exec_num))
        try:
            with self.get_file_text(info_file) as log_file:
                for line in log_file:
                    key, value = line.strip().split(':', 1)
                    if key in self.log:
                        self.log[key].append(value)
                    else:
                        self.log[key] = [value]
        except OSError as error:
            raise OSError(
                'Error while reading execution log file %s. %r' % (info_file, error)
            ) from error

    def get_execution_time(self) -> Optional[float]:
        """Return the time spent in the sandbox, reading the logs if
        necessary.

        return (float): time spent in the sandbox.

        """
        assert self.log is not None
        if 'time' in self.log:
            return float(self.log['time'][0])
        return None

    def get_execution_wall_clock_time(self) -> Optional[float]:
        """Return the total time from the start of the sandbox to the
        conclusion of the task, reading the logs if necessary.

        return (float): total time the sandbox was alive.

        """
        assert self.log is not None
        if 'time-wall' in self.log:
            return float(self.log['time-wall'][0])
        return None

    def use_soft_timeout(self) -> bool:
        return True

    def get_memory_used(self) -> Optional[int]:
        """Return the memory used by the sandbox, reading the logs if
        necessary.

        return (int): memory used by the sandbox (in kbytes).

        """
        assert self.log is not None
        if 'cg-mem' in self.log:
            # Isolate returns memory measurements in KiB.
            return int(self.log['cg-mem'][0])
        return None

    def get_killing_signal(self) -> int:
        """Return the signal that killed the sandboxed process,
        reading the logs if necessary.

        return (int): offending signal, or 0.

        """
        assert self.log is not None
        if 'exitsig' in self.log:
            return int(self.log['exitsig'][0])
        return 0

    def get_exit_code(self) -> int:
        """Return the exit code of the sandboxed process, reading the
        logs if necessary.

        return (int): exitcode, or 0.

        """
        assert self.log is not None
        if 'exitcode' in self.log:
            return int(self.log['exitcode'][0])
        return 0

    def get_status_list(self) -> List[str]:
        """Reads the sandbox log file, and set and return the status
        of the sandbox.

        return (list): list of statuses of the sandbox.

        """
        assert self.log is not None
        if 'status' in self.log:
            return self.log['status']
        return []

    def get_exit_status(self) -> str:
        """Get the list of statuses of the sandbox and return the most
        important one.

        return (string): the main reason why the sandbox terminated.

        """
        assert self.log is not None
        status_list = self.get_status_list()
        if 'XX' in status_list:
            return self.EXIT_SANDBOX_ERROR
        elif 'TO' in status_list:
            if 'message' in self.log and 'wall' in self.log['message'][0]:
                return self.EXIT_TIMEOUT_WALL
            else:
                return self.EXIT_TIMEOUT
        elif 'SG' in status_list:
            return self.EXIT_SIGNAL
        elif 'RE' in status_list:
            return self.EXIT_NONZERO_RETURN
        # OK status is not reported in the log file, it's implicit.
        return self.EXIT_OK

    def get_human_exit_description(self) -> str:
        """Get the status of the sandbox and return a human-readable
        string describing it.

        return (string): human-readable explaination of why the
                         sandbox terminated.

        """
        status = self.get_exit_status()
        if status == self.EXIT_OK:
            return (
                'Execution successfully finished (with exit code %d)'
                % self.get_exit_code()
            )
        elif status == self.EXIT_SANDBOX_ERROR:
            return 'Execution failed because of sandbox error'
        elif status == self.EXIT_TIMEOUT:
            return 'Execution timed out'
        elif status == self.EXIT_TIMEOUT_WALL:
            return 'Execution timed out (wall clock limit exceeded)'
        elif status == self.EXIT_SIGNAL:
            return 'Execution killed with signal %s' % self.get_killing_signal()
        elif status == self.EXIT_NONZERO_RETURN:
            return 'Execution failed because the return code was nonzero'
        return ''

    def inner_absolute_path(self, path: pathlib.Path) -> pathlib.Path:
        """Translate from a relative path inside the sandbox to an
        absolute path inside the sandbox.

        path (string): relative path of the file inside the sandbox.

        return (string): the absolute path of the file inside the sandbox.

        """
        return self._home_dest / path

    def _popen(
        self,
        command: List[str],
        stdin: Optional[IO[bytes] | int] = None,
        stdout: Optional[IO[bytes] | int] = None,
        stderr: Optional[IO[bytes] | int] = None,
        close_fds: bool = True,
    ) -> subprocess.Popen:
        """Execute the given command in the sandbox using
        subprocess.Popen, assigning the corresponding standard file
        descriptors.

        command ([string]): executable filename and arguments of the
            command.
        stdin (file|None): a file descriptor.
        stdout (file|None): a file descriptor.
        stderr (file|None): a file descriptor.
        close_fds (bool): close all file descriptor before executing.

        return (Popen): popen object.

        """
        self.log = None
        self.exec_num += 1

        # We run a selection of commands without isolate, as they need
        # to create new files. This is safe because these commands do
        # not depend on the user input.
        if command[0] in IsolateSandbox.SECURE_COMMANDS:
            logger.debug(
                'Executing non-securely: %s at %s',
                str(command),
                self._home,
            )
            try:
                prev_permissions = stat.S_IMODE(self._home.stat().st_mode)
                self._home.chmod(0o700)
                with open(self.cmd_file, 'at', encoding='utf-8') as cmds:
                    cmds.write('%s\n' % str(command))
                p = subprocess.Popen(
                    command,
                    cwd=str(self._home),
                    stdin=stdin,
                    stdout=stdout,
                    stderr=stderr,
                    close_fds=close_fds,
                )
                self._home.chmod(prev_permissions)
                # For secure commands, we clear the output so that it
                # is not forwarded to the contestants. Secure commands
                # are "setup" commands, which should not fail or
                # provide information for the contestants.
                if self.params.stdout_file:
                    (self._home / self.params.stdout_file).open('wb').close()
                if self.params.stderr_file:
                    (self._home / self.params.stderr_file).open('wb').close()
                self._write_empty_run_log(self.exec_num)
            except OSError:
                logger.critical(
                    'Failed to execute program in sandbox with command: %s',
                    str(command),
                    exc_info=True,
                )
                raise
            return p

        args = [self.box_exec] + self.build_box_options() + ['--'] + command
        logger.debug(
            "Executing program in sandbox with command: `%s'.",
            str(args),
        )
        # Temporarily allow writing new files.
        prev_permissions = stat.S_IMODE(self._home.stat().st_mode)
        self._home.chmod(0o700)
        with open(self.cmd_file, 'at', encoding='utf-8') as commands:
            commands.write('%s\n' % (str(args)))
        self._home.chmod(prev_permissions)
        try:
            p = subprocess.Popen(
                args, stdin=stdin, stdout=stdout, stderr=stderr, close_fds=close_fds
            )
        except OSError:
            logger.critical(
                'Failed to execute program in sandbox ' 'with command: %s',
                str(args),
                exc_info=True,
            )
            raise

        return p

    def _write_empty_run_log(self, index: int):
        """Write a fake run.log file with no information."""
        info_file = pathlib.PosixPath('%s.%d' % (self.info_basename, index))
        with info_file.open('wt', encoding='utf-8') as f:
            f.write('time:0.000\n')
            f.write('time-wall:0.000\n')
            f.write('max-rss:0\n')
            f.write('cg-mem:0\n')

    def execute_without_std(
        self,
        command: List[str],
    ) -> bool:
        """Execute the given command in the sandbox using
        subprocess.Popen and discarding standard input, output and
        error. More specifically, the standard input gets closed just
        after the execution has started; standard output and error are
        read until the end, in a way that prevents the execution from
        being blocked because of insufficient buffering.

        command ([string]): executable filename and arguments of the
            command.

        return (bool|Popen): return True if the sandbox didn't report
            errors (caused by the sandbox itself), False otherwise.
        """
        popen = self._popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
        )

        # If the caller wants us to wait for completion, we also avoid
        # std*** to interfere with command. Otherwise we let the
        # caller handle these issues.
        with popen as p:
            exitcode = self.translate_box_exitcode(
                wait_without_std([p], actually_pipe_to_stdout=self.debug)[0]
            )
        self.hydrate_logs()
        return exitcode

    def translate_box_exitcode(self, exitcode: int) -> bool:
        """Translate the sandbox exit code to a boolean sandbox success.

        Isolate emits the following exit codes:
        * 0 -> both sandbox and internal process finished successfully (meta
            file will contain "status:OK" -> return True;
        * 1 -> sandbox finished successfully, but internal process was
            terminated, e.g., due to timeout (meta file will contain
            status:x" with x in (TO, SG, RE)) -> return True;
        * 2 -> sandbox terminated with an error (meta file will contain
            "status:XX") -> return False.

        """
        if exitcode == 0 or exitcode == 1:
            return True
        elif exitcode == 2:
            return False
        else:
            raise Exception('Sandbox exit status (%d) unknown' % exitcode)

    def initialize(self):
        """Initialize isolate's box."""
        init_cmd = (
            [self.box_exec]
            + (['--cg'] if self.params.cgroup else [])
            + ['--box-id=%d' % self.box_id, '--init']
        )
        try:
            subprocess.check_call(init_cmd)
        except subprocess.CalledProcessError as e:
            raise Exception('Failed to initialize sandbox') from e

    def cleanup(self, delete: bool = False):
        """See Sandbox.cleanup()."""
        # The user isolate assigns within the sandbox might have created
        # subdirectories and files therein, making the user outside the sandbox
        # unable to delete the whole tree. If the caller asked us to delete the
        # sandbox, we first issue a chmod within isolate to make sure that we
        # will be able to delete everything. If not, we leave the files as they
        # are to avoid masking possible problems the admin wanted to debug.

        exe = (
            [self.box_exec]
            + (['--cg'] if self.params.cgroup else [])
            + ['--box-id=%d' % self.box_id]
        )

        if delete:
            # Ignore exit status as some files may be owned by our user
            subprocess.call(
                exe
                + [
                    '--dir=%s=%s:rw' % (str(self._home_dest), str(self._home)),
                    '--run',
                    '--',
                    '/bin/chmod',
                    '777',
                    '-R',
                    str(self._home_dest),
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
            )

        # Tell isolate to cleanup the sandbox.
        subprocess.check_call(
            exe + ['--cleanup'], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )

        if delete:
            logger.debug('Deleting sandbox in %s.', self._outer_dir)
            # Delete the working directory.
            shutil.rmtree(str(self._outer_dir))



import logging
import pathlib
import threading
import time
from typing import List, Optional

import fastapi
import jinja2
import rich
import rich.prompt
import rich.status
import uvicorn

from rbx import hydration, metadata, providers, utils
from rbx.config import Language, get_config
from rbx.console import console
from rbx.schema import DumpedProblem, Problem


def clear_loggers():
    for logger_name in [
        'uvicorn',
        'uvicorn.access',
        'uvicorn.asgi',
    ]:
        logging.getLogger(logger_name).handlers.clear()
        logging.getLogger(logger_name).propagate = False


def create_problem_structure(
    root: pathlib.Path,
    problem: Problem,
    lang: Language,
    status: Optional[rich.status.Status],
    should_simplify: bool = False,
    verbose: bool = False,
) -> Optional[DumpedProblem]:
    # Create directory structure.
    root.parent.mkdir(parents=True, exist_ok=True)

    problem_to_dump = DumpedProblem.from_problem(
        problem,
        code=providers.get_code(problem, simplify=should_simplify),
        aliases=providers.get_aliases(problem),
    )

    if verbose:
        console.print(
            f'Creating problem structure for [item]{problem_to_dump.pretty_name()}[/item]...'
        )

    code_path = root / lang.get_file(problem_to_dump.code)
    json_path = root / f'{problem_to_dump.code}.rbx.json'

    existing_problem = metadata.find_problem_by_code(problem_to_dump.code, root)
    if existing_problem:
        console.print(
            f'[error]Problem with identifier [item]{problem_to_dump.code}[/item] already exists in this folder.[/error]'
        )
        if not utils.confirm_on_status(
            status, 'Do you want to overwrite it?', default=False
        ):
            console.print(
                f'Skipping problem [item]{problem_to_dump.pretty_name()}[/item].'
            )
            return None

    json_path.write_text(utils.model_json(problem_to_dump))
    code = jinja2.Template(lang.get_template()).render(**problem_to_dump.get_vars())
    code_path.write_text(code)

    if verbose:
        console.print(
            f'Problem structure for [item]{problem_to_dump.pretty_name()}[/item] created successfully.'
        )
    return problem_to_dump


def process_problems(
    problems: List[Problem], lang: Language, status: rich.status.Status
):
    console.print(
        f'Creating problem structure for [item]{len(problems)}[/item] problems...'
    )

    should_simplify = False
    if providers.should_simplify_contest_problems(problems):
        console.print('Detected the parsed problems are from a contest.')
        if utils.confirm_on_status(
            status,
            'Do you want to identify these problems by their letters?',
            default=True,
        ):
            should_simplify = True

    root = pathlib.Path()
    dumped_problems = []
    for problem in problems:
        dumped_problem = create_problem_structure(
            root, problem, lang, status, should_simplify=should_simplify
        )
        if dumped_problem:
            dumped_problems.append(dumped_problem)
    console.print(f'Hydrating [item]{len(dumped_problems)}[/item] problems...')
    for problem in dumped_problems:
        hydration.hydrate_problem(root, problem)


def main(lang: Optional[str] = None):
    if get_config().get_language(lang) is None:
        console.print(
            f'[error]Language {lang or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
        )
        return

    app = fastapi.FastAPI()

    async def shutdown():
        server.should_exit = True

    batch_to_left_lock = threading.Lock()
    batch_to_left = {}
    ignored = set()
    saved_status = None
    problems_to_process = []

    def process_batch_item(problem: Problem):
        batch_to_left_lock.acquire()
        if problem.batch.id in ignored:
            batch_to_left_lock.release()
            return True
        if problem.batch.id not in batch_to_left:
            if len(batch_to_left) > 0:
                console.print(
                    f'[error]Ignoring extra batch [item]{problem.batch.id}[/item] since other batch is being parsed.[/error]'
                )
                ignored.add(problem.batch.id)
                batch_to_left_lock.release()
                return True
            if problem.batch.size > 1 and saved_status:
                saved_status.update(
                    f'[rbx]rbx[/rbx] is parsing problems from group [item]{problem.group}[/item]'
                )
            elif saved_status:
                saved_status.update('[rbx]rbx[/rbx] is parsing problems...')
            console.print(
                f'Started parsing batch [item]{problem.batch.id}[/item] with size [item]{problem.batch.size}[/item].'
            )
            batch_to_left[problem.batch.id] = problem.batch.size
        console.print(f'Parsing problem [item]{problem.name}[/item]...')
        problems_to_process.append(problem)
        finished = False
        if batch_to_left[problem.batch.id] == 1:
            finished = True
            if problem.batch.size > 1:
                console.print(
                    f'[status][rbx]rbx[/rbx] parsed all problems from group [item]{problem.group}[/item].[/status]'
                )
            else:
                console.print(
                    f'[status][rbx]rbx[/rbx] parsed problem from [item]{problem.url}[/item][/status]'
                )
        else:
            batch_to_left[problem.batch.id] -= 1
        batch_to_left_lock.release()
        return not finished

    clock = None

    @app.post('/')
    async def parse(problem: Problem, background_tasks: fastapi.BackgroundTasks):
        nonlocal clock
        if clock is None:
            clock = time.monotonic()
        if not process_batch_item(problem):
            duration = time.monotonic() - clock
            console.print(
                f'Parsed all problems in [item]{duration:.2f}[/item] seconds.'
            )
            background_tasks.add_task(shutdown)
        return {}

    config = uvicorn.Config(app, port=1327)
    server = uvicorn.Server(config=config)
    clear_loggers()
    with console.status('Waiting for Competitive Companion request...') as status:
        saved_status = status
        server.run()

    with console.status('Processing parsed problems...') as status:
        language = get_config().get_language(lang)
        if not language:
            console.print(
                f'[error]Language {lang or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
            )
            return
        process_problems(problems_to_process, language, status)



from rbx.box.packaging.polygon import xml_schema as polygon_schema

if __name__ == '__main__':
    print(
        polygon_schema.Problem.from_xml("""
<problem>
  <names>
    <name language="english" value="Poliedro"/>
    <name language="portuguese" value="Poliedro"/>
  </names>
  <statements>
    <statement charset="UTF-8" language="english" mathjax="true" path="statements/english/problem.tex" type="application/x-tex"/>
    <statement charset="UTF-8" language="portuguese" mathjax="true" path="statements/portuguese/problem.tex" type="application/x-tex"/>
    <statement charset="UTF-8" language="english" mathjax="true" path="statements/.html/english/problem.html" type="text/html"/>
    <statement charset="UTF-8" language="portuguese" mathjax="true" path="statements/.html/portuguese/problem.html" type="text/html"/>
    <statement language="english" path="statements/.pdf/english/problem.pdf" type="application/pdf"/>
    <statement language="portuguese" path="statements/.pdf/portuguese/problem.pdf" type="application/pdf"/>
  </statements>
  <judging input-file="" output-file="">
    <testset name="tests">
      <time-limit>1000</time-limit>
      <memory-limit>268435456</memory-limit>
      <test-count>13</test-count>
      <input-path-pattern>tests/%02d</input-path-pattern>
      <answer-path-pattern>tests/%02d.a</answer-path-pattern>
      <tests>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
        <test method="manual"/>
      </tests>
    </testset>
  </judging>
  <files>
    <resources>
      <file path="files/jngen.h" type="h.g++"/>
      <file path="files/olymp.sty"/>
      <file path="files/problem.tex"/>
      <file path="files/statements.ftl"/>
      <file path="files/testlib.h" type="h.g++"/>
    </resources>
  </files>
  <assets>
    <checker name="std::ncmp.cpp" type="testlib">
      <source path="files/check.cpp" type="cpp.g++17"/>
      <binary path="check.exe" type="exe.win32"/>
      <copy path="check.cpp"/>
    </checker>                              
  </assets>                                       
</problem>
""")
    )

    print(
        polygon_schema.Contest.from_xml("""
<contest>
<names>
<name language="russian" main="true" value="2012-2013 Тренировка СПбГУ B #12 Бинарный поиск, тернарный поиск"/>
</names>
<statements>
<statement language="russian" path="statements/russian/20122013-tryenirovka-spbgu-b-12-binarnyy-poisk-tyernarnyy-poisk-ru.pdf" type="application/pdf"/>
</statements>
<problems>
<problem index="A" path="problems/A"/>
<problem index="B" path="problems/B"/>
<problem index="C" path="problems/C"/>
<problem index="D" path="problems/D"/>
<problem index="E" path="problems/E"/>
</problems>
</contest>
""")
    )
